"""abstraction of khl concept channel: where messages flow in"""
import json
from abc import ABC, abstractmethod
from typing import Union, List, Dict

from . import api
from ._types import MessageTypes, ChannelTypes, SlowModeTypes, MessageFlagModes
from .gateway import Requestable, Gateway
from .interface import LazyLoadable
from .permission import ChannelPermission, PermissionHolder
from .role import Role
from .user import User, GuildUser
from .util import unpack_value, unpack_id


class Channel(LazyLoadable, Requestable, ABC):
    """
    Interface, represents a channel where messages flowing
    """
    type: ChannelTypes

    @abstractmethod
    async def send(self, content: Union[str, List], *, type: MessageTypes = None, **kwargs):
        """
        send a msg to the channel
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def id(self) -> str:
        """the channel's id

        this field should be protected, thus only exported a read-only prop"""
        raise NotImplementedError


class PublicChannel(Channel, PermissionHolder, ABC):
    """the channels in guild, in contrast to PrivateChannel(private chat)"""
    name: str
    user_id: str
    guild_id: str
    topic: str
    is_category: int
    parent_id: str
    level: int

    def __init__(self, **kwargs):
        self._id: str = kwargs.get('id')
        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_')
        self._update_fields(**kwargs)

    @property
    def id(self) -> str:
        return self._id

    def _update_fields(self, **kwargs):
        self.name: str = kwargs.get('name')
        self.user_id: str = kwargs.get('user_id')
        self.guild_id: str = kwargs.get('guild_id')
        self.topic: str = kwargs.get('topic')
        self.is_category: int = kwargs.get('is_category')
        self.parent_id: str = kwargs.get('parent_id')
        self.level: int = kwargs.get('level')
        self.type: ChannelTypes = kwargs.get('type') and ChannelTypes(kwargs.get('type'))
        self.permission: ChannelPermission = ChannelPermission(**kwargs)

    async def load(self):
        self._update_fields(**(await self.gate.exec_req(api.Channel.view(self.id))))
        self._loaded = True

    async def update(self, name: str = None, topic: str = None, slow_mode: Union[int, SlowModeTypes] = None) -> dict:
        """
        update channel's settings
        """
        params = {'channel_id': self.id}
        if name is not None:
            params['name'] = name
        if topic is not None:
            params['topic'] = topic
        if slow_mode is not None:
            params['slow_mode'] = unpack_value(slow_mode)
        rt = await self.gate.exec_req(api.Channel.update(**params))
        await self.load()
        return rt

    async def list_users(self,
                         search: str = None,
                         role: Union[Role, str, int] = None,
                         mobile_verified: bool = None,
                         active_time: int = None,
                         joined_at: int = None,
                         page: int = 1,
                         page_size: int = 50,
                         filter_user_id: str = None) -> List[User]:
        """list the users who can see this channel"""
        params = {'guild_id': self.guild_id, 'channel_id': self.id, 'page': page, 'page_size': page_size}
        if search is not None:
            params['search'] = search
        if role is not None:
            params['role_id'] = unpack_id(role)
        if mobile_verified is not None:
            params['mobile'] = 1 if mobile_verified else 0
        if active_time is not None and active_time in [0, 1]:
            params['active_time'] = active_time
        if joined_at is not None and joined_at in [0, 1]:
            params['joined_at'] = joined_at
        if filter_user_id is not None:
            params['filter_user_id'] = filter_user_id
        users = await self.gate.exec_paged_req(api.Guild.userList(**params))
        return [User(_gate_=self.gate, _lazy_loaded_=True, **i) for i in users]

    async def list_messages(self,
                            page_size: int = None,
                            pin: int = None,
                            flag: MessageFlagModes = None,
                            msg_id: str = None) -> Dict:
        """list the messages in this channel (only for public channel now)"""
        params = {'target_id': self.id}
        if page_size is not None:
            params['page_size'] = page_size
        if pin is not None:
            params['pin'] = pin
        if flag is not None:
            params['flag'] = flag
        if msg_id is not None:
            params['msg_id'] = msg_id
        return await self.gate.exec_req(api.Message.list(**params))


class PublicTextChannel(PublicChannel):
    """
    `Standard Object`

    Text chat channels in guild
    """
    slow_mode: int

    def _update_fields(self, **kwargs):
        super()._update_fields(**kwargs)
        self.slow_mode: int = kwargs.get('slow_mode')

    async def send(self, content: Union[str, List], *, type: MessageTypes = None, temp_target_id: str = '', **kwargs):
        """
        send a msg to the channel

        ``temp_target_id`` is available in PublicTextChannel, so ``send()`` is overloaded here
        """
        # if content is card msg, then convert it to plain str
        if isinstance(content, List):
            type = MessageTypes.CARD
            content = json.dumps(content)
        type = type if type is not None else MessageTypes.KMD

        # merge params
        kwargs['target_id'] = self.id
        kwargs['content'] = content
        kwargs['type'] = type.value
        if temp_target_id:
            kwargs['temp_target_id'] = temp_target_id

        return await self.gate.exec_req(api.Message.create(**kwargs))


class PublicVoiceChannel(PublicChannel):
    """
    Voice chat channel in guild

    a placeholder now for future design/adaption
    """

    async def send(self, content: Union[str, List], **kwargs):
        raise TypeError('now there is no PublicVoiceChannel, *hey dude we have a pkg from future*')

    async def move_user(self, *users: Union[User, str]):
        """move users to this public voice channel"""
        user_ids = [u.id if isinstance(u, User) else u for u in users]
        return await self.gate.exec_req(api.Channel.moveUser(target_id=self.id, user_ids=user_ids))

    async def fetch_user_list(self) -> List[GuildUser]:
        """get users chatting in the voice channel"""
        users = await self.gate.exec_req(api.Channel.userList(channel_id=self.id))
        return [GuildUser(_gate_=self.gate, guild_id=self.guild_id, _lazy_loaded_=True, **i) for i in users]


def public_channel_factory(_gate_: Gateway, **kwargs) -> Union[PublicTextChannel, PublicVoiceChannel]:
    """factory function to build a channel object"""
    kwargs['type'] = kwargs['type'] if isinstance(kwargs['type'], ChannelTypes) else ChannelTypes(kwargs['type'])
    if kwargs['type'] == ChannelTypes.TEXT:
        return PublicTextChannel(**kwargs, _gate_=_gate_)
    if kwargs['type'] == ChannelTypes.VOICE:
        return PublicVoiceChannel(**kwargs, _gate_=_gate_)
    raise ValueError(f'unsupported channel type: {kwargs["type"]}: {kwargs}')


class PrivateChannel(Channel):
    """Private chat channel

    a private channel associates the code and another user(called as target in the following)"""

    code: str
    last_read_time: int
    latest_msg_time: int
    unread_count: int
    is_friend: bool
    is_blocked: bool
    is_target_blocked: bool
    target_info: Dict

    def __init__(self, **kwargs):
        self.code: str = kwargs.get('code')
        self.last_read_time: int = kwargs.get('last_read_time')
        self.latest_msg_time: int = kwargs.get('latest_msg_time')
        self.unread_count: int = kwargs.get('unread_count')
        self.is_friend: bool = kwargs.get('is_friend')
        self.is_blocked: bool = kwargs.get('is_blocked')
        self.is_target_blocked: bool = kwargs.get('is_target_blocked')
        self.target_info: Dict = kwargs.get('target_info')

        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_')

    async def load(self):
        pass

    @property
    def id(self) -> str:
        return self.target_user_id

    @property
    def target_user_id(self) -> str:
        """prop, the target's id"""
        return self.target_info.get('id') if self.target_info else None

    @property
    def target_user_name(self) -> str:
        """prop, the target's name"""
        return self.target_info.get('username') if self.target_info else None

    @property
    def is_target_user_online(self) -> bool:
        """prop, is the target online"""
        return self.target_info.get('online') if self.target_info else None

    @property
    def target_user_avatar(self) -> str:
        """prop, the target's avatar"""
        return self.target_info.get('avatar') if self.target_info else None

    async def send(self, content: Union[str, List], *, type: MessageTypes = None, **kwargs):
        return await User(id=self.id, _gate_=self.gate).send(content, type=type, **kwargs)



from ._types import RoleTypes


class Role:
    """
    `Standard Object`

    represent the component that used in permission control and user identify
    """
    id: int
    name: str
    color: int
    position: int
    hoist: int
    mentionable: int
    permissions: int
    type: RoleTypes

    def __init__(self, **kwargs):
        self.id: int = kwargs.get("role_id", 0)
        self.name: str = kwargs.get("name", "")
        self.color: int = kwargs.get("color", 0)
        self.position: int = kwargs.get("position", 0)
        self.hoist: int = kwargs.get("hoist", 0)
        self.mentionable: int = kwargs.get("mentionable", 0)
        self.permissions: int = kwargs.get("permissions", 0)
        self.type: RoleTypes = RoleTypes(kwargs.get("type", 0))

    def has_permission(self, bit_value: int) -> bool:
        """
        bitValue see https://developer.kaiheila.cn/doc/http/guild-role#%E6%9D%83%E9%99%90%E8%AF%B4%E6%98%8E
        """
        return self.permissions & (1 << bit_value) != 0



import re
import datetime
from typing import Union, Dict, List

from .struct import Struct
from .element import Element
from .interface import Types, _Module

_RE_INVITE_CODE = re.compile(r'^[a-zA-Z0-9]{6}$')
_RE_INVITE_LINK = re.compile(r'^https://kaihei\.co/[a-zA-Z0-9]{6}$')


class Module:
    """docs: https://github.com/kaiheila/api-docs/blob/main/docs/zh-cn/cardmessage.md#%E6%A8%A1%E5%9D%97"""

    class Header(_Module):
        """please refer to module docs"""
        _type = 'header'
        _text: Element.Text

        def __init__(self, text: Union[Element.Text, str] = ''):
            if isinstance(text, str):
                text = Element.Text(text)
            # Force header to use plain because KOOK right now only supports using plain in header
            self._text = Element.Text(text.content, Types.Text.PLAIN, text.emoji)
            super().__init__(Types.Theme.NA, Types.Size.NA)

        @property
        def text(self) -> Element.Text:
            """contained text"""
            return self._text

        @text.setter
        def text(self, value: Union[Element.Text, str]):
            self._text = Element.Text(value) if isinstance(value, str) else value

        @property
        def _repr(self) -> Union[Dict, str]:
            return self._gen_dict(['type', 'text'])

    class Section(_Module):
        """docs: https://t.ly/uM4K"""
        _type = 'section'
        _text: Element.Text
        _accessory: Union[Element.Image, Element.Button, None]
        mode: Types.SectionMode

        def __init__(self,
                     text: Union[Element.Text, str, Struct.Paragraph] = '',
                     accessory: Union[Element.Image, Element.Button, None] = None,
                     mode: Union[Types.SectionMode, str] = Types.SectionMode.LEFT):
            if isinstance(text, str):
                text = Element.Text(text)
            self.text = text
            self.mode = mode
            self.accessory = accessory
            super().__init__(Types.Theme.NA, Types.Size.NA)

        @property
        def text(self) -> Element.Text:
            """contained text"""
            return self._text

        @text.setter
        def text(self, value: Union[Element.Text, str]):
            self._text = Element.Text(value) if isinstance(value, str) else value

        @property
        def mode(self) -> Types.SectionMode:
            """the section's mode, refet to SectionMode for enum detail"""
            return self._mode

        @mode.setter
        def mode(self, value: Union[Types.SectionMode, str]):
            self._mode = Types.SectionMode(value) if isinstance(value, str) else value

        @property
        def accessory(self) -> Union[Element.Image, Element.Button]:
            """the accessory attached to the section, a button/an image"""
            return self._accessory

        @accessory.setter
        def accessory(self, value: Union[Element.Image, Element.Button]):
            if isinstance(value, Element.Button):
                self.mode = Types.SectionMode.RIGHT
            self._accessory = value

        @property
        def _repr(self) -> Union[Dict, str]:
            return self._gen_dict(['type', 'text', 'mode', 'accessory'])

    class ImageGroup(_Module):
        """docs: https://t.ly/2yaa"""
        _type = 'image-group'
        _elements: List[Element.Image]

        def __init__(self, *images: Element.Image):
            if not 1 <= len(images) <= 9:
                raise ValueError('element length unacceptable, should: 9 >= len >= 1')
            self._elements = list(images)
            super().__init__(Types.Theme.NA, Types.Size.NA)

        def append(self, image: Element.Image):
            """append a image into"""
            if len(self._elements) >= 9:
                raise ValueError('element max length exceeded(9)')
            self._elements.append(image)

        def pop(self, index: int = ...):
            """pop a image from"""
            if len(self._elements) <= 1:
                raise ValueError('element min length exceeded(1)')
            return self._elements.pop(index)

        def len(self) -> int:
            """count of current elements"""
            return len(self._elements)

        @property
        def _repr(self) -> Union[Dict, str]:
            return self._gen_dict(['type', 'elements'])

    class ActionGroup(_Module):
        """a group of buttons(now only support buttons)

        docs: https://t.ly/LBVJ"""
        _type = 'action-group'
        _elements: List[Element.Button]

        def __init__(self, *elements: Element.Button):
            self._elements = list(elements)
            super().__init__(Types.Theme.NA, Types.Size.NA)

        def append(self, element: Element.Button):
            """append a button into"""
            self._elements.append(element)

        def pop(self, index: int = None) -> Element.Button:
            """pop a button from"""
            return self._elements.pop(index)

        @property
        def _repr(self) -> Dict:
            return self._gen_dict(['type', 'elements'])

    class Context(_Module):
        """a set of text/images

        docs: https://t.ly/_ECm"""
        _type = 'context'
        _elements: List[Union[Element.Text, Element.Image]]

        def __init__(self, *elements: Union[Element.Text, Element.Image, str]):
            self._elements = [Element.Text(i) if isinstance(i, str) else i for i in elements]
            super().__init__(Types.Theme.NA, Types.Size.NA)

        def append(self, element: Union[Element.Text, Element.Image, str]):
            """append a text/image into"""
            self._elements.append(Element.Text(element) if isinstance(element, str) else element)

        def pop(self, index: int = None):
            """pop a text/image from"""
            return self._elements.pop(index)

        @property
        def _repr(self) -> Dict:
            return self._gen_dict(['type', 'elements'])

    class Divider(_Module):
        """docs: https://t.ly/vpuy"""
        _type = 'divider'

        def __init__(self):
            super().__init__(Types.Theme.NA, Types.Size.NA)

        @property
        def _repr(self) -> Dict:
            return {'type': 'divider'}

    class Invite(_Module):
        """component to invite others to a server

        docs: https://t.ly/q-5V"""
        _type = 'invite'

        def __init__(self, code: str = ''):
            if _RE_INVITE_CODE.match(code) or _RE_INVITE_LINK.match(code):
                self._code = code
            else:
                raise ValueError('invite code or invite link is not valid')
            super().__init__(Types.Theme.NA, Types.Size.NA)

        @property
        def code(self) -> str:
            """invite code"""
            return self._code

        @code.setter
        def code(self, value: str):
            if _RE_INVITE_CODE.match(value) or _RE_INVITE_LINK.match(value):
                self._code = value
            else:
                raise ValueError('invite code or invite link is not valid')

        @property
        def _repr(self) -> Dict:
            return self._gen_dict(['type', 'code'])

    class File(_Module):
        """show file in a module

        docs: https://t.ly/U9tm"""
        src: str
        title: str
        cover: str

        def __init__(self, type: Union[Types.File, str], src: str, title: str = '', cover: str = ''):
            self._type = type.value if isinstance(type, Types.File) else type
            self.src = src
            self.title = title
            self.cover = cover
            super().__init__(Types.Theme.NA, Types.Size.NA)

        @property
        def _repr(self) -> Dict:
            d = self._gen_dict(['type', 'src', 'title'])
            if self._type == Types.File.AUDIO.value and self.cover:
                d['cover'] = self.cover
            return d

    class Countdown(_Module):
        """count down module

        docs: https://t.ly/Xwtp"""
        _type = "countdown"
        end: datetime.datetime
        mode: Types.CountdownMode
        start: datetime.datetime

        def __init__(self,
                     end: datetime.datetime,
                     *,
                     mode: Union[Types.CountdownMode, str] = Types.CountdownMode.HOUR,
                     start: datetime.datetime = None):
            self.end = end
            self.mode = mode if isinstance(mode, Types.CountdownMode) else Types.CountdownMode(mode)
            self.start = start
            super().__init__(Types.Theme.NA, Types.Size.NA)

        @property
        def _repr(self) -> Dict:
            d = self._gen_dict(['type', 'mode'])
            if self.mode == Types.CountdownMode.SECOND and self.start:
                d['startTime'] = int(self.start.timestamp() * 1000)
            d['endTime'] = int(self.end.timestamp() * 1000)
            return d

    class Container(ImageGroup):
        """contains images, like image-group but the images are not stripped into square

        docs: https://t.ly/Rpn6"""
        _type = 'container'



import inspect
import logging
from typing import Any, Dict, Callable, Coroutine

from khl import Message

log = logging.getLogger(__name__)

TypeEHandler = Callable[['Command', Exception, Message], Coroutine]


class Exceptions:
    """built-in exceptions raised from command process procedure"""

    class Lexer:
        """built-in exceptions raised from lexer"""

        class Base(Exception):
            """base of all kinds of exceptions raised from lexer"""

        class NotMatched(Base):
            """the msg content is not matched to the command"""

        class LexFailed(Base):
            """exception raised during lex process"""

            def __init__(self, exc, *args):
                """
                :param exc: original exception raised from lexer
                """
                self.exc = exc
                super().__init__(*args)

    class Parser:
        """built-in exceptions raised from parser"""

        class Base(Exception):
            """base of all kinds of exceptions raised from parser"""

        class NoParseFunc(Base):
            """there is no parse function for expected_type"""

            def __init__(self, expected_param: inspect.Parameter, token: str, *args):
                """
                :param expected_param: expected param info
                :param token: given token
                """
                self.expected_param = expected_param
                self.token_value = token
                super().__init__(*args)

        class ParseFailed(Base):
            """exception raised during token parser function execution"""

            def __init__(self, expected_param: inspect.Parameter, token: str, parse_func, exc, *args):
                """
                :param expected_param: expected param info
                :param token: given token
                :param parse_func: parse function used to parse the token
                :param exc: original exception raised by parse_func
                """
                self.expected_param = expected_param
                self.token = token
                self.parse_func = parse_func
                self.exc = exc
                super().__init__(*args)

    class Handler:
        """built-in exceptions raised during command handling"""

        class Base(Exception):
            """base of all kinds of exceptions raised during command handling"""

        class RuleNotPassed(Base):
            """rules associated with the command not passed"""

            def __init__(self, rule, *args):
                self.rule = rule
                super().__init__(*args)

        class ArgLenNotMatched(Base):
            """arg list length is not matched, too many args/not enough args"""

            def __init__(self, expected_min: int, expected_max: int, actual: int, *args):
                """

                :param expected_min: min expected arg nums
                :param expected_max: max expected arg nums
                :param actual: actual given arg nums
                """
                self.expected_min = expected_min
                self.expected_max = expected_max
                self.actual = actual
                super().__init__(*args)


async def log_on_exc(cmd, e, _):
    """basically log the exception"""
    if isinstance(e, Exceptions.Lexer.NotMatched):  # ignore NotMatched since it is kinda noisy
        return
    log.exception(f'exception raised when handling command: {cmd.name}', exc_info=e)


async def ignore_exc(*_):
    """explicitly ignore the exception"""
    return


default_exc_handler: Dict[Any, TypeEHandler] = {
    Exception: log_on_exc,  # log in default
}



"""helper wrapper for color in card message"""
import re
from typing import Tuple, Union, Optional

from .interface import _Representable


class Color(_Representable):
    """abstraction of color, provides helper functions"""

    def __init__(self, *rgb: int, hex_color: str = None):
        if (not rgb or len(rgb) != 3) and not hex_color:
            raise ValueError('rgb(as a tuple) or hex required')
        if hex_color:
            match = re.match(r'^#?([\da-fA-F]{2})([\da-fA-F]{2})([\da-fA-F]{2})$', hex_color)
            if not match:
                raise ValueError('unacceptable hex color')
            self._r, self._g, self._b = (int(match.group(i), 16) for i in (1, 2, 3))
        else:
            self._r, self._g, self._b = (self._rgb_check(i) for i in rgb)

    @staticmethod
    def _rgb_check(value: int) -> int:
        if not 0 <= value <= 255:
            raise ValueError(f'unacceptable rgb value, expected [0,255], exact {value}')
        return value

    @property
    def r(self) -> int:
        """red channel component in rgb model"""
        return self._r

    @r.setter
    def r(self, value: int):
        Color._rgb_check(value)
        self._r = value

    @property
    def g(self) -> int:
        """green channel component in rgb model"""
        return self._g

    @g.setter
    def g(self, value: int):
        Color._rgb_check(value)
        self._g = value

    @property
    def b(self) -> int:
        """blue channel component in rgb model"""
        return self._b

    @b.setter
    def b(self, value: int):
        Color._rgb_check(value)
        self._b = value

    def hex(self) -> str:
        """hex string form"""
        return self._repr

    @property
    def _repr(self) -> str:
        return f'#{self._r:02x}{self._g:02x}{self._b:02x}'


def make_color(color: Union[Color, Tuple[int, int, int], str, None]) -> Optional[Color]:
    """helper to unify all forms of color"""
    result = None
    if isinstance(color, Color):
        result = color
    elif isinstance(color, tuple):
        result = Color(*color)
    elif isinstance(color, str):
        result = Color(hex_color=color)
    return result



from abc import ABC, abstractmethod
from typing import List, Union

from . import api, util
from .role import Role
from .user import User
from .interface import LazyLoadable
from .gateway import Requestable


class RolePermission:
    """part of channel permission, the permission setting for a specific role in the channel

    this setting overwrites the role's permission set in the guild"""
    role_id: int
    allow: int
    deny: int

    def __init__(self, *, role_id: int, allow: int, deny: int, **_):
        self.role_id = role_id
        self.allow = allow
        self.deny = deny


class UserPermission:
    """part of channel permission, the permission setting for a specific user"""
    user: User
    allow: int
    deny: int

    def __init__(self, *, user: User, allow: int, deny: int, **_):
        self.user = user
        self.allow = allow
        self.deny = deny


class ChannelPermission(LazyLoadable, Requestable):
    """permission settings in a channel, which can be customized.

    the custom permission entry can be divided into two types:
    1. customized role permission: overwrites the role's global permission settings
    2. customized user permission: exclusively set a user's permission in the channel
    """
    _id: str  # bound channel id
    _sync: int

    roles: List[RolePermission]
    users: List[UserPermission]

    @property
    def id(self) -> str:
        """which channel the permission belongs to"""
        return self._id

    @property
    def sync(self) -> bool:
        """if this channel's permission sync with category"""
        return self._sync != 0

    @sync.setter
    def sync(self, value: bool):
        self._sync = 1 if value else 0

    def __init__(self, **kwargs):
        self._id: str = kwargs.get('id')
        self.gate = kwargs.get('_gate_')
        self._load_fields(**kwargs)

    def _load_fields(self, **kwargs):
        self.roles = [RolePermission(**i) for i in kwargs.get('permission_overwrites', [])]
        self.users = [UserPermission(**i) for i in kwargs.get('permission_users', [])]
        self._sync = kwargs.get('permission_sync', None)
        if self.roles and self.users and (self._sync is not None):
            self._loaded = True

    async def load(self):
        self._load_fields(**await self.gate.exec_req(api.ChannelRole.index(channel_id=self.id)))


class PermissionHolder(LazyLoadable, Requestable, ABC):
    """holder of the permissions, can be a channel or a category"""

    permission: ChannelPermission

    @property
    @abstractmethod
    def id(self) -> str:
        """the channel's id

        this field should be protected, thus only exported a read-only prop"""
        raise NotImplementedError

    async def fetch_permission(self, force_update: bool = True) -> ChannelPermission:
        """fetch permission setting of the channel"""
        if force_update or not self.permission.loaded:
            await self.permission.load()
        return self.permission

    async def create_user_permission(self, target: Union[User, str]):
        """create a customized permission setting entry"""
        t = 'user_id'
        v = util.unpack_id(target)
        d = await self.gate.exec_req(api.ChannelRole.create(channel_id=self.id, type=t, value=v))
        self.permission.loaded = False
        return d

    async def update_user_permission(self, target: Union[User, str], allow: int = 0, deny: int = 0) -> Role:
        """update a customized permission setting entry"""
        t = 'user_id'
        v = util.unpack_id(target)
        return await self.gate.exec_req(
            api.ChannelRole.update(channel_id=self.id, type=t, value=v, allow=allow, deny=deny))

    async def delete_user_permission(self, target: Union[User, str]):
        """delete a customized permission setting entry"""
        t = 'user_id'
        v = util.unpack_id(target)
        return await self.gate.exec_req(api.ChannelRole.delete(channel_id=self.id, type=t, value=v))

    async def create_role_permission(self, target: Union[Role, str]):
        """create a customized permission setting entry"""
        t = 'role_id'
        v = util.unpack_id(target)
        d = await self.gate.exec_req(api.ChannelRole.create(channel_id=self.id, type=t, value=v))
        self.permission.loaded = False
        return d

    async def update_role_permission(self, target: Union[Role, str], allow: int = 0, deny: int = 0) -> Role:
        """update a customized permission setting entry"""
        t = 'role_id'
        v = util.unpack_id(target)
        return await self.gate.exec_req(
            api.ChannelRole.update(channel_id=self.id, type=t, value=v, allow=allow, deny=deny))

    async def delete_role_permission(self, target: Union[Role, str]):
        """delete a customized permission setting entry"""
        t = 'role_id'
        v = util.unpack_id(target)
        return await self.gate.exec_req(api.ChannelRole.delete(channel_id=self.id, type=t, value=v))



"""card message component: Card"""
from typing import List, Union, Tuple, Optional, Dict

from .color import Color, make_color
from .interface import Types, _Common
from .module import _Module


class Card(_Common):
    """abstraction of card, which contains modules"""
    _type = 'card'
    _color: Optional[Color]
    _modules: List[_Module]

    def __init__(self,
                 *modules: _Module,
                 color: Union[Color, Tuple[int, int, int], str, None] = None,
                 theme: Union[Types.Theme, str, None] = None,
                 size: Union[Types.Size, str, None] = Types.Size.LG):
        self._modules = list(modules)
        self._color = make_color(color)
        super().__init__(theme, size)

    def append(self, module: _Module):
        """append a module into the card"""
        self._modules.append(module)

    def pop(self, index: int = None) -> _Module:
        """pop the `index`th module from the card"""
        return self._modules.pop(index)

    @property
    def color(self) -> Optional[Color]:
        """color of the card's frame"""
        return self._color

    @color.setter
    def color(self, value: Union[Color, Tuple[int, int, int], str]):
        self._color = make_color(value)

    @property
    def theme(self) -> Types.Theme:
        """the card's theme, when color is not filled, the card's frame will be set according to theme"""
        return self._theme

    @theme.setter
    def theme(self, value: Union[Types.Theme, str]):
        self._theme = Types.Theme(value)

    @property
    def size(self) -> Types.Size:
        """the size of the card"""
        return self._size

    @size.setter
    def size(self, value: Union[Types.Size, str]):
        self._size = Types.Size(value)

    @property
    def _repr(self) -> Dict:
        return self._gen_dict(['type', 'theme', 'size', 'color', 'modules'])



from typing import Union, List, Dict

from .element import Element
from .interface import _Struct, Types


class Struct:
    """component type: struct

    doc: https://developer.kaiheila.cn/doc/cardmessage#%E7%BB%93%E6%9E%84%E4%BD%93
    """

    class Paragraph(_Struct):
        """component: Struct.Paragraph"""
        _type = 'paragraph'
        _cols: int
        _fields: List[Element.Text]

        def __init__(self, cols: int, *fields: Union[Element.Text, str]):
            if not 1 <= len(fields) <= 50:
                raise ValueError('fields length unacceptable, should: 50 >= len >= 1')
            self._cols = cols
            self._fields = [f if isinstance(f, Element.Text) else Element.Text(f) for f in fields]
            super().__init__(Types.Theme.NA, Types.Size.NA)

        def append(self, field: Element.Text):
            """append a text element to the paragraph

            maximal size: 50, raise ValueError if exceeded"""
            if len(self._fields) >= 50:
                raise ValueError('fields max length exceeded(50)')
            self._fields.append(field)

        def pop(self, index: int) -> Element.Text:
            """pop the index th text element from the paragraph"""
            if len(self._fields) <= 1:
                raise ValueError('fields min length exceeded(1)')
            return self._fields.pop(index)

        def len(self):
            """current length of self text elements"""
            return len(self._fields)

        @property
        def _repr(self) -> Union[str, Dict]:
            return self._gen_dict(['type', 'cols', 'fields'])



import json
from typing import List, Union

from . import api
from ._types import MessageTypes, FriendTypes
from .gateway import Requestable, Gateway
from .interface import LazyLoadable
from .intimacy import Intimacy
from .role import Role


class User(LazyLoadable, Requestable):
    """
    `Standard Object`

    represent a entity that interact with khl server
    """
    id: str
    username: str
    nickname: str
    identify_num: str
    online: bool
    bot: bool
    status: int
    avatar: str
    vip_avatar: str
    mobile_verified: bool

    _loaded: bool
    gate: Gateway

    def __init__(self, **kwargs):
        self.id = kwargs.get('id', '')
        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_', None)
        self._update_fields(**kwargs)

    def _update_fields(self, **kwargs):
        self.username = kwargs.get('username', '')
        self.nickname = kwargs.get('nickname', '')
        self.identify_num = kwargs.get('identify_num', '')
        self.online = kwargs.get('online', False)
        self.bot = kwargs.get('bot', False)
        self.status = kwargs.get('status', 0)
        self.avatar = kwargs.get('avatar', '')
        self.vip_avatar = kwargs.get('vip_avatar', '')
        self.mobile_verified = kwargs.get('mobile_verified', False)

    async def load(self):
        self._update_fields(**(await self.gate.exec_req(api.User.view(self.id))))
        self._loaded = True

    async def send(self, content: Union[str, List], *, type: MessageTypes = None, **kwargs):
        """
        send a msg to a channel

        ``temp_target_id`` is only available in ChannelPrivacyTypes.GROUP
        """
        # if content is card msg, then convert it to plain str
        if isinstance(content, List):
            type = MessageTypes.CARD
            content = json.dumps(content)
        else:
            type = type or MessageTypes.KMD

        # merge params
        kwargs['target_id'] = self.id
        kwargs['content'] = content
        kwargs['type'] = type.value

        return await self.gate.exec_req(api.DirectMessage.create(**kwargs))

    async def fetch_intimacy(self) -> Intimacy:
        """get the user's intimacy info"""
        return Intimacy(user_id=self.id, **(await self.gate.exec_req(api.Intimacy.index(user_id=self.id))))

    async def update_intimacy(self, score: int = 0, social_info: str = None, img_id: str = None):
        """update the user's intimacy"""
        params = {'user_id': self.id}
        if score is not None:
            params['score'] = score
        if social_info is not None:
            params['social_info'] = social_info
        if img_id is not None:
            params['img_id'] = img_id
        return await self.gate.exec_req(api.Intimacy.update(**params))

    async def add_friend(self):
        """send friend request to the user"""
        await self.gate.exec_req(api.Friend.request(user_code=f'{self.username}#{self.identify_num}', _from=0))

    async def block(self):
        """block the user"""
        await self.gate.exec_req(api.Friend.block(user_id=self.id))


class GuildUser(User):
    """a user in guild

    with some fields more than User"""
    guild_id: str
    joined_at: int
    active_time: int
    roles: List[int]
    gate: Gateway

    def _update_fields(self, **kwargs):
        super()._update_fields(**kwargs)
        self.roles = kwargs.get('roles', [])
        self.guild_id = kwargs.get('guild_id', '')
        self.joined_at = kwargs.get('joined_at', 0)
        self.active_time = kwargs.get('active_time', 0)

    async def load(self):
        self._update_fields(**(await self.gate.exec_req(api.User.view(self.id, self.guild_id))))
        self._loaded = True

    async def fetch_roles(self, **kwargs) -> List[Role]:
        """
        Get the user roles in this guild

        paged req, support standard pagination args

        :return: A list for Role
        """
        guild_roles = (await self.gate.exec_paged_req(api.GuildRole.list(self.guild_id), **kwargs))
        rt: List[Role] = []
        for role in guild_roles:
            if role['role_id'] in self.roles:
                rt.append(Role(**role))
        return rt

    async def set_nickname(self, nickname: str):
        """
        Set user's nickname
        """
        await self.gate.exec_req(api.Guild.nickname(guild_id=self.guild_id, nickname=nickname, user_id=self.id))

    async def add_friend(self):
        await self.gate.exec_req(
            api.Friend.request(user_code=f'{self.username}#{self.identify_num}', _from=2, guild_id=self.guild_id))


class Friend:
    """
    Friend with specific friend id and friend info
    """

    gate: Gateway

    id: int
    user_id: str

    _user: User
    _type: FriendTypes

    def __init__(self, _gate_: Gateway, **kwargs):
        self.gate = _gate_
        self.id = kwargs.get('id')
        self.user_id = kwargs.get('user_id')
        self._type = kwargs.get('type')

    async def fetch_user(self) -> User:
        """get user"""
        if self._user is None:
            self._user = User(_gate_=self.gate, **(await self.gate.exec_req(api.User.view(user_id=self.user_id))))
        return self._user

    async def delete(self):
        """delete the friend"""
        await self.gate.exec_req(api.Friend.delete(user_id=self.user_id))

    async def block(self):
        """block the user"""
        await self.gate.exec_req(api.Friend.block(user_id=self.user_id))

    async def unblock(self):
        """unblock the blocked user"""
        await self.gate.exec_req(api.Friend.unblock(user_id=self.user_id))

    @property
    def type(self) -> FriendTypes:
        """the type of the friend"""
        return self._type


class FriendRequest(Friend):
    """
    Friend request with specific id and user info
    """

    async def accept(self):
        """accept the friend request"""
        await self.gate.exec_req(api.Friend.handleRequest(id=self.id, accept=1))
        return Friend(_gate_=self.gate, user_id=self.user_id)

    async def deny(self):
        """deny the friend request"""
        await self.gate.exec_req(api.Friend.handleRequest(id=self.id, accept=0))

    @property
    def type(self) -> FriendTypes:
        return FriendTypes.REQUEST



import asyncio
import logging
import time
import zlib
from abc import ABC, abstractmethod
from typing import Dict

from aiohttp import ClientWebSocketResponse, ClientSession, web, WSMessage

from .cert import Cert
from .interface import AsyncRunnable

log = logging.getLogger(__name__)

API = 'https://www.kaiheila.cn/api/v3'


class Receiver(AsyncRunnable, ABC):
    """
    1. receive raw data from khl server
    2. decrypt & parse raw data into pkg
    3. put pkg into the pkg_queue() for others to use
    """
    _queue: asyncio.Queue

    @property
    def type(self) -> str:
        """the network type used by the receiver"""
        raise NotImplementedError

    @property
    def pkg_queue(self) -> asyncio.Queue:
        """output port of the receiver"""
        return self._queue

    @pkg_queue.setter
    def pkg_queue(self, queue: asyncio.Queue):
        self._queue = queue

    @abstractmethod
    async def start(self):
        """run self"""
        raise NotImplementedError


class WebsocketReceiver(Receiver):
    """receive data in websocket mode"""

    def __init__(self, cert: Cert, compress: bool):
        super().__init__()
        self._cert = cert
        self.compress = compress

        self._NEWEST_SN = 0
        self._RAW_GATEWAY = ''

    @property
    def type(self) -> str:
        return 'websocket'

    async def heartbeat(self, ws_conn: ClientWebSocketResponse):
        """khl customized heartbeat scheme"""
        while True:
            try:
                await asyncio.sleep(26)
                await ws_conn.send_json({'s': 2, 'sn': self._NEWEST_SN})
            except ConnectionResetError:
                return
            except Exception as e:
                log.exception('error raised during websocket heartbeat',
                              exc_info=e)

    async def _get_gateway(self, cs: ClientSession):
        headers = {
            'Authorization': f'Bot {self._cert.token}',
            'Content-type': 'application/json'
        }
        params = {'compress': 1 if self.compress else 0}
        async with cs.get(f"{API}/gateway/index",
                          headers=headers,
                          params=params) as res:
            res_json = await res.json()
            if res_json['code'] != 0:
                log.error(f'getting gateway: {res_json}')
                return

            self._RAW_GATEWAY = res_json['data']['url']

    async def _connect_gateway_and_handle_msg(self, cs: ClientSession):
        async with cs.ws_connect(self._RAW_GATEWAY) as ws_conn:
            asyncio.ensure_future(self.heartbeat(ws_conn), loop=self.loop)

            log.info('[ init ] launched')
            try:
                async for raw in ws_conn:
                    raw: WSMessage
                    await self._handle_raw(raw)
            except Exception:
                log.exception(
                    'error raised during websocket receive, reconnect automatically'
                )

    async def start(self):
        async with ClientSession(loop=self.loop) as cs:
            while True:
                await self._get_gateway(cs)
                await self._connect_gateway_and_handle_msg(cs)

    async def _handle_raw(self, raw: WSMessage):
        try:
            data = raw.data
            data = zlib.decompress(data) if self.compress else data
            pkg: Dict = self._cert.decode_raw(data)
            log.debug(f'upcoming raw: {pkg}')
            if pkg['s'] != 0:
                return
            self._NEWEST_SN = pkg['sn']
            await self.pkg_queue.put(pkg['d'])
        except Exception as e:
            log.exception(e)


class WebhookReceiver(Receiver):
    """receive data in webhook mode"""

    def __init__(self, cert: Cert, *, port: int, route: str, compress: bool):
        super().__init__()
        self._cert = cert
        self.port = port
        self.route = route
        self.app = web.Application()
        self.compress = compress
        self.sn_dup_map = {}

    @property
    def type(self) -> str:
        return 'webhook'

    def _is_dup(self, req: dict) -> bool:
        sn = req.get('sn', None)
        if sn is None:
            return False
        current = time.time()
        if sn in self.sn_dup_map:
            if current - self.sn_dup_map[sn] <= 600:
                # 600 sec timed out
                return True
        self.sn_dup_map[sn] = current
        return False

    async def start(self):

        async def on_recv(request: web.Request):
            try:
                data = await request.read()
                data = zlib.decompress(data) if self.compress else data
                pkg: Dict = self._cert.decode_raw(data)
            except Exception as e:
                log.exception(e)
                return web.Response()

            if not pkg:  # empty pkg
                return web.Response()

            if pkg['d'][
                    'verify_token'] != self._cert.verify_token:  # check verify_token
                return web.Response()

            if self._is_dup(pkg):  # dup pkg
                return web.Response()

            if pkg['s'] == 0:
                pkg = pkg['d']
                if pkg['type'] == 255 and pkg[
                        'channel_type'] == 'WEBHOOK_CHALLENGE':
                    return web.json_response({'challenge': pkg['challenge']})
                await self.pkg_queue.put(pkg)

            return web.Response()

        self.app.router.add_post(self.route, on_recv)

        runner = web.AppRunner(self.app)
        await runner.setup()  # runner use its own loop, can not be set
        site = web.TCPSite(runner, '0.0.0.0', self.port)

        log.info('[ init ] launched')

        await site.start()

        while True:
            await asyncio.sleep(3600)  # sleep forever



"""ABC"""
import asyncio
from abc import ABC, abstractmethod


class AsyncRunnable(ABC):
    """
    Classes that has async work to run
    """
    _loop: asyncio.AbstractEventLoop = None

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """the event loop for the async work"""
        return self._loop

    @loop.setter
    def loop(self, new_loop: asyncio.AbstractEventLoop):
        self._loop = new_loop

    def schedule(self):
        """schedule the async work into background"""
        asyncio.ensure_future(self.start(), loop=self.loop)

    @abstractmethod
    async def start(self):
        """run the async work"""


class LazyLoadable(ABC):
    """
    Classes that can be initialized before loaded full data from khl server.
    These classes objects usually are constructed by khl.py internal calls.

    For example:
        `Channel`: we usually construct a channel with a message for convenient,
        while we only know the channel's id, so this channel is not `loaded`, until call the `load()`
    """
    _loaded: bool

    @abstractmethod
    async def load(self):
        """
        Load full data from khl server

        :return: empty
        """
        raise NotImplementedError

    @property
    def loaded(self) -> bool:
        """check if loaded"""
        return self._loaded

    @loaded.setter
    def loaded(self, v: bool):
        self._loaded = v

    def is_loaded(self) -> bool:
        """DEPRECATED: use ``.loaded`` prop, simpler in code and clearer in semantic
        Check if loaded

        :return: bool
        """
        return self._loaded



from typing import List, Tuple


class Intimacy:
    """the user's intimacy info"""
    user_id: str
    img_url: str
    social_info: str
    last_read: int
    score: int
    img_list: List[Tuple[str, str]]

    def __init__(self, **kwargs) -> None:
        self.user_id = kwargs.get('user_id')
        self.img_url = kwargs.get('img_url')
        self.social_info = kwargs.get('social_info')
        self.last_read = kwargs.get('last_read')
        self.score = kwargs.get('score')
        self.img_list = [(img.get('id'), img.get('url')) for img in kwargs.get('img_list')]



"""card

card is a message type in khl, which provides kinds of expressive components to convey"""
from .card import Card
from .card_message import CardMessage
from .color import Color
from .element import Element
from .interface import Types
from .module import Module
from .struct import Struct



"""game: components used in playing status"""
from typing import List


class Game:
    """represents a game used in playing state"""
    id: int
    name: str
    type: int
    options: str
    kmhook_admin: bool
    process_name: List[str]
    product_name: List[str]
    icon: str

    def __init__(self, **kwargs) -> None:
        self.id = kwargs.get('id', '')
        self.name = kwargs.get('name', '')
        self.type = kwargs.get('type', 1)
        self.options = kwargs.get('options', '')
        self.product_name = kwargs.get('product_name', [])
        self.icon = kwargs.get('icon', '')

    def __repr__(self):
        return f'Game(id={self.id}, name="{self.name}", type={self.type}, options="{self.options}",' \
               f' product_name={self.product_name}, icon="{self.icon}")'



"""component Command"""
import asyncio
import inspect
import logging
from copy import copy
from typing import Callable, Coroutine, List, Union, Pattern, Any, Dict

from khl import Message, Client
from .exception import default_exc_handler, TypeEHandler, Exceptions
from .lexer import Lexer, RELexer, DefaultLexer
from .parser import Parser
from .rule import TypeRule
from .util import wrap_if_coro

log = logging.getLogger(__name__)

TypeHandler = Callable[..., Coroutine]


def _get_arg_by_key_type(annotation, predefined_args: dict):
    for v in predefined_args.values():
        if isinstance(v, annotation):
            return v
    return None


class Command:
    """
    Receive a `Message`, parse it with `lex` and `parser`, execute `handler` if successfully parsed else discard it
    """
    name: str
    handler: TypeHandler
    help: str
    desc: str

    lexer: Lexer
    parser: Parser

    rules: List[TypeRule]
    exc_handlers: Dict[Any, TypeEHandler]

    def __init__(
        self,
        name: str,
        handler: TypeHandler,
        help: str,
        desc: str,
        lexer: Lexer,
        parser: Parser,
        rules: List[TypeRule],
        exc_handlers: Dict[Any, TypeEHandler]
    ):
        if not asyncio.iscoroutinefunction(handler):
            raise TypeError('handler must be a coroutine.')
        self.handler = handler

        self.name = name or handler.__name__
        if not isinstance(self.name, str):
            raise TypeError('Name of a command must be a string.')

        self.help = help
        self.desc = desc

        self.lexer = lexer
        self.parser = parser

        self.rules = list(rules)
        self.exc_handlers = copy(default_exc_handler) if exc_handlers is None else dict(exc_handlers)

    @staticmethod
    def command(
        name: str = '',
        *,
        help: str = '',
        desc: str = '',
        aliases: List[str] = (),
        prefixes: List[str] = ('/', ),
        regex: Union[str, Pattern] = '',
        lexer: Lexer = None,
        parser: Parser = None,
        rules: List[TypeRule] = (),
        exc_handlers: Dict[Any, TypeEHandler] = None,
        case_sensitive: bool = True
    ) -> Callable[[TypeHandler], 'Command']:
        """
        decorator, to wrap a func into a Command

        :param name: the name of this Command, also used to trigger command in DefaultLexer
        :param aliases: (DefaultLexer only) you can also trigger the command with aliases
        :param prefixes: (DefaultLexer only) command prefix, default use '/'
        :param regex: (RELexer only) pattern for the command
        :param help: detailed manual
        :param desc: short introduction
        :param lexer: (Advanced) explicitly set the lexer
        :param parser: (Advanced) explicitly set the parser
        :param rules: command executed if all rules are checked
        :param case_sensitive: is the command trigger sensitive
        :return: a decorator to wrap Command
        """
        if not lexer and regex:
            lexer = RELexer(regex) if isinstance(regex, str) else regex

        def decorator(handler: TypeHandler):
            default_lexer = DefaultLexer(set(prefixes), set([name or handler.__name__] + list(aliases)), case_sensitive)
            return Command(name, handler, help, desc, lexer or default_lexer, parser or Parser(), rules, exc_handlers)

        return decorator

    def on_exception(self, type=Exception):
        """returns a decorator, register an exception handler to the command

        handler is invoked when exception/subclass exception raised during command handling

        the exception handler signature & usage example::

            @bot.command()
            async def foo(m: Message, i: int, *a: float):
                ...

            @foo.on_exception(Exceptions.Handler.ArgLenNotMatched)
            async def foo_on_arg_len(cmd: Command, exc: Exceptions.Handler.ArgLenNotMatched, msg: Message):
                ...

        note: the param `exc` can be typed according to the type passed to `on_exception`
        """

        def decorator(handler: TypeEHandler):
            self.exc_handlers[type] = handler

        return decorator

    async def handle(self, msg: Message, client: Client, predefined_kwargs: dict):
        """handle msg:

        1. check if matched the command
        2. if matched, execute the command handler"""
        try:
            predefined_args, to_be_parsed = self._split_params(predefined_kwargs)
            parsed_args = await self.parser.parse(msg, client, self.lexer.lex(msg), to_be_parsed)

            log.info(f'command {self.name} is triggered by msg: {msg.content}')

            await self._check_rules(msg)
            self._check_arg_len(to_be_parsed, parsed_args)

            await self.handler(*predefined_args, *parsed_args)
        except Exception as e:
            return await self._handle_exc(e, msg)

    def _split_params(self, predefined_kwargs: dict) -> (List[Any], List[inspect.Parameter]):
        """get parameters that need to be parsed according to `ignores`

        :param predefined_kwargs: list of predefined kwargs

        :return: a tuple: (the used predefined args, the params need to be parsed)
        """
        predefined_args = []
        to_be_parsed = []
        params = list(inspect.signature(self.handler).parameters.values())
        for p in params:
            predefined = _get_arg_by_key_type(p.annotation, predefined_kwargs)
            if predefined is not None:
                predefined_args.append(predefined)
            else:
                to_be_parsed.append(p)
        return predefined_args, to_be_parsed

    async def _check_rules(self, msg: Message):
        """:raise Exceptions.Handler.RuleNotPassed: if there is rule not passed"""
        # usually, there is only one or two rules, thus no need to check rules concurrently
        for rule in self.rules:
            if not bool(await wrap_if_coro(rule(msg))):
                raise Exceptions.Handler.RuleNotPassed(rule)

    @staticmethod
    def _check_arg_len(to_be_parsed: List[inspect.Parameter], parsed_args: List):
        """:raise Exceptions.Handler.ArgLenNotMatched: if parsed_args len not matched to_be_parsed"""

        def is_positional(p):
            return p.kind is inspect.Parameter.VAR_POSITIONAL

        def has_default(p):
            return p.default is not inspect.Parameter.empty

        minn = sum(map(
            lambda p: 1 if not is_positional(p) and not has_default(p) else 0,
            to_be_parsed,
        ))
        maxx = -1 if inspect.Parameter.VAR_POSITIONAL in map(lambda x: x.kind, to_be_parsed) else len(to_be_parsed)
        if len(parsed_args) < minn or (maxx != -1 and len(parsed_args) > maxx):
            raise Exceptions.Handler.ArgLenNotMatched(minn, maxx, len(parsed_args))

    async def _handle_exc(self, e, msg: Message):
        """lookup self.exc_handlers if type(e) exists, execute it"""
        handlers = [self.exc_handlers[h] for h in self.exc_handlers if isinstance(e, h)]
        return await asyncio.gather(*[h(self, e, msg) for h in handlers])
